#ifndef COMPUTE_TREE_HH
# define COMPUTE_TREE_HH

#include <mln/border/fill.hh>
#include <mln/border/resize.hh>

#include "immerse.hh"
#include "propagation.hh"
#include "max_tree.hh"
#include "canon.hh"

namespace mln
{
    namespace K2
    {
        template <typename T>
            void
            fill_borders_shift(image2d< T > &F, T val)
            {
                typedef image2d<T> I;
                typedef T* ptr_i;

                const unsigned nrows = F.nrows(), ncols = F.ncols();
                int offset_s = F.delta_index(dpoint2d(1, 0));
                ptr_i cell;

                /*
                 * horizontal
                 */
                cell = &F.at_(-1, -1);
                for(unsigned i = 0; i < ncols; i++)
                    *(cell + i) = val;

                /*
                 * vertical
                 */
                cell = &F.at_(-1, ncols - 2);
                for(unsigned i = 1; i < nrows - 1; i++)
                    *(cell + i * offset_s) = val;

                /*
                 * horizontal
                 */
                cell = &F.at_(nrows - 2, -1);
                for(unsigned i = 0; i < ncols; i++)
                    *(cell + i) = val;

                /*
                 * vertical
                 */
                cell = &F.at_(-1, -1);
                for(unsigned i = 1; i < nrows - 1; i++)
                    *(cell + i * offset_s) = val;
            }

        /// Returns the ToS as a `parent` relationship between each point.
        /// A parent of (-1, -1) is assumed to be a point outside of the image.
        image2d<point2d> compute_tree(image2d< value::int_u<8u> >& f)
        {
            unsigned l_start;
            iqueue_fast< unsigned, value::int_u<9> > q;

            typedef irange< value::int_u<9> > R;

            // immerse
            image2d<R> F = K2::irange_immerse_opt_median_queue(f, l_start, q, true);

            // compute the tree
            const box2d& b = F.domain();

            int offset_n = F.delta_index(dpoint2d(-1, 0));
            int offset_s = F.delta_index(dpoint2d(1, 0));

            // images buffers
            image2d<unsigned> seen(b);
            border::resize(seen, 1);

            // size
            q.init(l_start, true);

            // initial values
            data::fill(seen, 0);
            border::fill(seen, 1);
            fill_borders_shift(seen, 2u);

            propagation_info* common = new propagation_info(
                    F.buffer(),
                    seen,
                    seen.buffer(),
                    offset_n,
                    offset_s,
                    F.nrows() + 2);
            propagation_thread_specific_info* root = new propagation_thread_specific_info(common);
            root->current = std::move(q);
            Worker* rootw = new(tbb::task::allocate_root()) Worker(root, 0, 2);
            tbb::task::spawn_root_and_wait(*rootw);
            SubtreeWorker sw(seen);
            tbb::parallel_reduce(tbb::blocked_range<size_t>(0, F.nrows()), sw);

            canon_first_pass(sw, F.ncols());
            canon_first_pass(sw, F.ncols());
            canon_second_pass(sw, F.ncols());

            image2d<point2d>  parent(f.domain());

            for (auto i = sw.getChildsRows().begin(); i != sw.getChildsRows().end(); ++i) {
                tbb::parallel_for(tbb::blocked_range<size_t>(0, F.ncols()),
                        [&](const tbb::blocked_range<size_t>& r) {
                            for (auto j = *i + r.begin(); j != *i + r.end(); ++j) {
                                if (K2::is_primary(j->pos)) {
                                    auto target = j->pos / 4;
                                    auto par    = j->par->pos / 4;

                                    parent(target) = K2::is_primary(j->par->pos) ? par : point2d(-1, -1);
                                }
                            }
                        });
            }

            return std::move(parent);
        } // compute_tree
    } // end of namespace mln::K2
} // end of namespace mln


#endif // ! COMPUTE_TREE_HH
