#ifndef IQUEUE_PARALLEL_HH
# define IQUEUE_PARALLEL_HH

# include <cassert>
# include <ostream>

# include "irange.hh"

namespace mln
{
  template <typename E, typename T>
    struct h_queue_fast
    {
      enum {
        nbits_  = mln_nbits(T),
        sup_ = mlc_pow_int(2, nbits_)
      };

      h_queue_fast()
      {
        size_ = 0;
      }

      void push(const E& e, const T& v)
      {
        q_[v].push(e);
        ++size_;
      }

      void swap_levels(int in, int out)
      {
        std::swap(q_[in], q_[out]);
      }

      E pop(const T& v)
      {
        mln_precondition(! is_empty(v));
        E e = q_[v].top();
        q_[v].pop();
        --size_;
        return e;
      }

      unsigned size() const
      {
        return size_;
      }

      bool is_empty() const
      {
        return size_ == 0;
      }

      bool is_empty(int v) const
      {
        assert(v >= 0 && v < sup_);
        return q_[v].empty();
      }

      void debug_println(std::ostream& ostr = std::cout) const
      {
        ostr << "[ ";
        for (unsigned v = 0; v < sup_; ++v)
        {
          if (q_[v].empty())
            continue;
          ostr << v << '-' << q_[v].size() << ' ';
        }
        ostr << "]  #" << size_ << std::endl;
      }

      private:

      std::stack<E> q_[sup_];
      unsigned size_;
    };




  template <typename E, typename T>
    struct iqueue_fast
    {

      enum {
        l_sup_ = h_queue_fast<E,T>::sup_
      };

      void init(const T& l, bool up)
      {
        l_ = l;
        up_ = up;
      }

      unsigned size() const
      {
        return q_.size();
      }

      bool direction() const
      {
        return up_;
      }

      void set_direction_down ()
      {
        up_ = false;
      }

      bool is_empty() const
      {
        return size() == 0;
      }

      bool is_level_done() const
      {
        return q_.is_empty(l_);
      }

      int level() const
      {
        return l_;
      }

      void set_level(int l) const
      {
        l_ = l;
      }

      void swap_levels(int in, int out)
      {
        q_.swap_levels(in, out);
      }

      T push_at_current_level(const E e)
      {
        q_.push(e, l_);
        return l_;
      }

      T push_at_level(const E e, const T v) {
        q_.push(e, v);
        return v;
      }

      E pop()
      {
        assert(! is_empty());

        if (! q_.is_empty(l_))
          return q_.pop(l_);

        if (up_)
        {
          for (; l_ < l_sup_; ++l_)
            if (! q_.is_empty(l_))
              return q_.pop(l_);
        }
        else // down
        {
          for (; l_ >= 0; --l_)
            if (! q_.is_empty(l_))
              return q_.pop(l_);
        }

        std::cerr << "(" << size() << ") weird: not empty but could not find one"
                  << std::endl;
        std::abort();

        return E();
      }

      void debug_println(std::ostream& ostr = std::cout) const
      {
        ostr << "l = " << l_ << " going " << (up_ ? "up" : "down") << ":"
             << std::endl;
        ostr << "  q_ = ";
        q_.debug_println(ostr);
      }

      bool up() const
      {
        return up_;
      }

      private:

      h_queue_fast<E,T> q_;
      bool up_;
      int l_;

      bool invariant_() const
      {
        if (up_)
        {
          for (int l = 0; l < l_; ++l)
            if (! q_.is_empty(l))
              return false;
        }
        else // down
        {
          for (int l = l_; l >= 0; --l)
            if (! q_.is_empty(l))
              return false;
        }
        return true;
      }
    };
} // end of namespace mln


#endif // ! IQUEUE_PARALLEL_HH
