#ifndef IRANGE_HH
# define IRANGE_HH

#include <mln/accu/internal/base.hh>

namespace mln
{
  // irange
  template <typename T>
  struct irange : Value< irange<T> >
  {
    typedef irange enc;
    typedef irange equiv;

    T lower, upper;

    typedef T value;

    irange()
    {
    }

    irange(T lower, T upper)
      : lower(lower),
      upper(upper)
    {
    }

    template <typename U>
    irange(const U& v)
      : lower(v),
      upper(v)
    {
    }

    template <typename U>
    irange<T>& operator=(const U& v)
    {
      lower = v;
      upper = v;
      return *this;
    }



    template <typename U>
    operator irange<U>() const
    {
      irange<U> i;
      i.lower = lower;
      i.upper = upper;
      return i;
    }



    unsigned size() const { return upper - lower + 1; }

    bool is_degenerated() const { return size() == 1; }

    irange<T> openmin() const
    {
      if (size() == 2) {
        std::cout << "unable to openmin irange " << *this << std::endl;
        std::abort();
      }

      if (is_degenerated())
        return *this;
      else
        return irange(lower + 1, upper);
    }

    irange<T> open() const
    {
      if (size() == 2) {
        std::cout << "unable to open irange " << *this << std::endl;
        std::abort();
      }

      if (is_degenerated())
        return *this;
      else
        return irange(lower + 1, upper - 1);
    }
    

    bool has(const T& v) const
    {
      return lower <= v && v <= upper;
    }

    bool invariant_() const { return upper >= lower; }


    void set_empty() {}
  };



  template <typename T>
  inline
  std::ostream&
  operator<<(std::ostream& ostr, const irange<T>& i)
  {
    if (i.is_degenerated())
      return ostr << '{' << i.lower << '}';
    else
      return ostr << "[|" << i.lower << ',' << i.upper << "|]";
  }


  template <typename T>
  inline
  irange<T>
  make_irange(T lower, T upper)
  {
    if (upper < lower) {
      std::cout << "iconsistant irange [|" << lower << "," << upper << "|]"
                << std::endl;
      std::abort();
    }
    irange<T> i(lower, upper);
    return i;
  }


  // thresholding

  template <typename T, typename U>
  inline
  bool
  operator<=(const irange<T>& i, const U& v)
  {
    return i.lower <= v;
  }

  template <typename T, typename U>
  inline
  bool
  operator>(const irange<T>& i, const U& v)
  {
    return ! (i <= v);
  }

  template <typename T, typename U>
  inline
  bool
  operator>=(const irange<T>& i, const U& v)
  {
    return v <= i.upper;
  }

  template <typename T, typename U>
  inline
  bool
  operator<(const irange<T>& i, const U& v)
  {
    return ! (i >= v);
  }

  template <typename T, typename U>
  inline
  bool
  operator==(const irange<T>& i, const U& v)
  {
    return i >= v && i <= v;
  }



  // comparisons

  template <typename T>
  inline
  bool
  operator==(const irange<T>& lhs, const irange<T>& rhs)
  {
    return lhs.lower == rhs.lower && lhs.upper == rhs.upper;
  }

  template <typename T>
  inline
  bool
  operator!=(const irange<T>& lhs, const irange<T>& rhs)
  {
    return ! (lhs == rhs);
  }
  

  // set operations


  template <typename T>
  inline
  irange<T>
  span(const irange<T>& lhs, const irange<T>& rhs)
  {
    irange<T> i(std::min(lhs.lower, rhs.lower),
		std::max(lhs.upper, rhs.upper));
    return i;
  }



  template <typename T>
  inline
  bool
  intersect(const irange<T>& lhs, const irange<T>& rhs)
  {
    return span(lhs, rhs).size() < lhs.size() + rhs.size();
  }



  template <typename T>
  inline
  irange<T>
  inter(const irange<T>& lhs, const irange<T>& rhs)
  {
    if (! intersect(lhs, rhs)) {
      std::cout << "Unable to intersect iranges" << std::endl;
      std::abort();
    }

    irange<T> i(std::max(lhs.lower, rhs.lower),
        std::min(lhs.upper, rhs.upper));
    return i;
  }




  // specific set operations


  template <typename T>
  inline
  bool
  adjacent(const irange<T>& lhs, const irange<T>& rhs)
  {
    return span(lhs, rhs).size() <= lhs.size() + rhs.size();
  }



  template <typename T>
  inline
  irange<T>
  cup(const irange<T>& lhs, const irange<T>& rhs)
  {
    if (! adjacent(lhs, rhs))
    {
      std::cerr << lhs << ' ' << rhs << std::endl;
      std::abort();
    }
    return span(lhs, rhs);
  }
} // mln

#endif // ! IRANGE_HH
