#ifndef MAX_TREE_HH
# define MAX_TREE_HH

#include <list>
#include <tbb/tbb.h>
#include <mln/core/image/image2d.hh>

#include "ufind_data.hh"

namespace mln
{
    namespace K2
    {
        class SubtreeWorker {
            unsigned                 _upper_row_i;
            unsigned                 _lower_row_i;
            std::list<ufind_data*>   _childs_rows;
            const image2d<unsigned>& _ord;
            unsigned                 _ncols;

            void build_1Dtree_for_line(ufind_data* row) {
                // build 1D max tree
                ufind_data* r = row;
                std::stack<ufind_data*> s;
                ufind_data* end = row + _ncols;
                for (ufind_data* p = row + 1; p != end; ++p) {
                    if (r->value < p->value) {
                        s.push(r);
                        r = p;
                    } else if (r->value == p->value)
                        p->par = r;
                    else {
                        while(true) {
                            ufind_data* q = nullptr;
                            if (!s.empty())
                                q = s.top();
                            if (q == nullptr || q->value < p->value) {
                                r->par = p;
                                r = p;
                                break;
                            } else if (q->value == p->value) {
                                r->par = q;
                                p->par = q;
                                r = q;
                                s.pop();
                                break;
                            } else {
                                r->par = q;
                                r = q;
                                s.pop();
                            }
                        }
                    }
                }

                while (!s.empty()) {
                    r->par = s.top();
                    r = s.top();
                    s.pop();
                }
                r->par = r; // root points on itself
            }

            ufind_data* levroot(ufind_data* x) const {
                if (x->par != x && x->value == x->par->value) {
                    x->par = levroot(x->par);
                    return x->par;
                }
                return x;
            }

            void connect_points(ufind_data* x, ufind_data* y) {
                x = levroot(x);
                y = levroot(y);
                if (y->value > x->value)
                    std::swap(x, y);
                while (x != y) {
                    if (x->par == x) {
                        x->par = y;
                        x = y;
                    } else {
                        ufind_data* z = levroot(x->par);
                        if (z->value > y->value)
                            x = z;
                        else {
                            x->par = y;
                            x = y;
                            y = z;
                        }
                    }
                }
            }

            void merge_1Dlines(ufind_data* up, ufind_data* down) {
                ufind_data* endup = up + _ncols;
                for (; up != endup; ++up, ++down)
                    connect_points(up, down);
            }

            public:
            const std::list<ufind_data*>& getChildsRows() const
            { return _childs_rows; }

            void operator()(const tbb::blocked_range<size_t>& subrange) {
                if (!_childs_rows.empty()) {
                    // this may be called with concecutive subrange (concecutive wrt the
                    // last operator() or join call.
                    SubtreeWorker sw(*this);
                    sw(subrange);
                    join(sw);
                    return;
                }
                tbb::cache_aligned_allocator<ufind_data> alloc;
                ufind_data* lastline;
                ufind_data* currline = nullptr;

                for (size_t i = subrange.begin(); i != subrange.end(); ++i) {
                    lastline = currline;
                    currline = alloc.allocate(_ncols);
                    ufind_data* end = currline + _ncols;

                    for (ufind_data* curr = currline; curr != end; ++curr) {
                        curr->pos   = point2d(i - 1, curr - currline - 1);
                        curr->value = _ord(curr->pos);
                    }

                    build_1Dtree_for_line(currline);

                    if (lastline != nullptr)
                        merge_1Dlines(lastline, currline);

                    _childs_rows.push_back(currline);
                }

                _upper_row_i = subrange.end() - 1;
                _lower_row_i = subrange.begin();
            }

            SubtreeWorker(const image2d<unsigned>& ord)
                : _ord(ord), _ncols(ord.ncols()) {
                }

            SubtreeWorker(const SubtreeWorker& other)
                : _ord(other._ord), _ncols(other._ncols) {
                }
            SubtreeWorker(SubtreeWorker& other, tbb::split)
                : _ord(other._ord), _ncols(other._ncols) {
                }

            void join (SubtreeWorker& other) {
                // merge two set of lines’ fronteers
                if (_upper_row_i > other._lower_row_i) {
                    merge_1Dlines(_childs_rows.front(), other._childs_rows.back());
                    _childs_rows.splice(_childs_rows.begin(), other._childs_rows);
                    _upper_row_i = other._upper_row_i;
                } else {
                    merge_1Dlines(_childs_rows.back(), other._childs_rows.front());
                    _childs_rows.splice(_childs_rows.end(), other._childs_rows);
                    _lower_row_i = other._lower_row_i;
                }
            }
        };
    }
}

#endif
