#ifndef PROPAGATION_HH_
# define PROPAGATION_HH_

#include "ufind_data.hh"

#include <mln/core/image/image2d.hh>

#include <tbb/tbb.h>

#include "iqueue_parallel.hh"
#include "irange.hh"

namespace mln
{
    namespace K2
    {
        // informations not duplicated for the parallelization process
        struct propagation_info {
            irange< value::int_u<9u> >* F;
            image2d<unsigned>&          seen_img;
            unsigned*                   seen;
            int                         offset_n;
            int                         offset_s;
            unsigned                    nrows;

            propagation_info(
                    irange< value::int_u<9u> >* _F,
                    image2d<unsigned>&          _seen_img,
                    unsigned*                   _seen,
                    int                         _offset_n,
                    int                         _offset_s,
                    unsigned                    _nrows
                    )
                : F(_F),
                seen_img(_seen_img),
                seen(_seen),
                offset_n(_offset_n),
                offset_s(_offset_s),
                nrows(_nrows)
            { }
        };

        struct propagation_thread_specific_info {
            propagation_info*                          infos;
            // This is actually the only value thread-local:
            iqueue_fast< unsigned, value::int_u<9u> >  current;

            propagation_thread_specific_info(propagation_info* _infos)
                : infos(_infos)
            { }
        };

        class Worker : public tbb::task {
            propagation_thread_specific_info* infos;
            int                               local_depth;
            unsigned                          baseord;

            public:
            Worker(propagation_thread_specific_info* _infos,
                    int                               _local_depth,
                    int                               _baseord)
                : infos(_infos), local_depth(_local_depth),
                baseord(_baseord) { }

            tbb::task* execute() {
                iqueue_fast< unsigned, value::int_u<9u> >& currqueue = infos->current;
                set_ref_count(1);

                // some data must be shared :
                // infos->in_queue, infos->K, infos->F, infos->seen,
                // infos->ord, infos->show

                // all passes: images offsets
                int offset_n = infos->infos->offset_n;
                int offset_s = infos->infos->offset_s;

                // images buffers
                irange< value::int_u<9u> >* F_base = infos->infos->F;
                unsigned*                   seen_base = infos->infos->seen;
                const int dp[] = { offset_n, offset_s, -1, 1 };

                // 1st pass: sorting
                //-------------------
                {
                    // propagation informations for the children
                    propagation_thread_specific_info* other;
                    other = new propagation_thread_specific_info(infos->infos);
                    unsigned int p;
                    unsigned     currord = baseord;
                    unsigned     bestlevel = currqueue.level();
                    while (!currqueue.is_empty())
                    {
                        p = currqueue.pop();
                        *(seen_base + p) = currord;

                        // first element
                        // unsigned int first_F = p + offset_n - 1;
                        for (unsigned i = 0u; i < 4u; ++i)
                        {
                            unsigned int n_F = p + dp[i];
                            if (!*(seen_base + n_F))
                            {
                                /* push:
                                 *   in the current queue if = current_level
                                 *   in the up      queue if > current_level
                                 *   in the down    queue if < current_level
                                 * Note: Because we are doing these tests now, it is not
                                 * needed
                                 * to recheck it in the queues' code.
                                 */
                                if (currqueue.level() > (int)(F_base + n_F)->upper)
                                { // new pixel value is smaller than the current level
                                    if (currqueue.up()) {
                                        other->current.push_at_level(n_F, (F_base + n_F)->upper);
                                        if (bestlevel < (F_base + n_F)->upper)
                                            bestlevel = (F_base + n_F)->upper;
                                    }
                                    else
                                        currqueue.push_at_level(n_F, (F_base + n_F)->upper);
                                }
                                else if (currqueue.level() < (F_base + n_F)->lower)
                                { // new pixel value is greater than the current level
                                    if (currqueue.up())
                                        currqueue.push_at_level(n_F, (F_base + n_F)->lower);
                                    else {
                                        other->current.push_at_level(n_F, (F_base + n_F)->lower);
                                        if (bestlevel > (F_base + n_F)->lower)
                                            bestlevel = (F_base + n_F)->lower;
                                    }
                                }
                                else // ((F_base + n_F)->has(currqueue.level()))
                                    currqueue.push_at_current_level(n_F);
                                *(seen_base + n_F) = 1;
                            }
                        }

                        if (currqueue.is_level_done()) {
                            ++currord;
                            if (other->current.size()) {
                                // The current flat surface has been extracted.
                                // We must:
                                //  - if both queues are empty: we are infos->done…
                                //  - if both queues are not empty: process the current queue on
                                //  this thread, and the other queue on a newly started thread.
                                other->current.init(bestlevel, !currqueue.up());
                                Worker* w = new(allocate_additional_child_of(task::self()))
                                    Worker(other, local_depth + 1, currord);
                                spawn(*w);
                                other = new propagation_thread_specific_info(infos->infos);
                                bestlevel = currqueue.level();
                            }
                        }
                    } // end of while
                    delete other; // will be an unused structure
                } // end of 1st pass
                wait_for_all();
                return 0; // no next task to execute
            }
        };

    }
}

#endif
